#ifndef TAINT_MARKS_H
#define TAINT_MARKS_H

#ifdef DARWIN
#include <sys/xattr.h>
#define fgetxattr(fd, name, value, size) \
	fgetxattr(fd, name, value, size, 0, 0)
#define fsetxattr(fd, name, value, size, flags) \
	fsetxattr(fd, name, value, size, 0, flags)
#define fremovexattr(fd, name) \
	fremovexattr(fd, name, 0)
#elif defined(__FreeBSD__)
#include <sys/extattr.h>
#define fgetxattr(fd, name, value, size) \
	extattr_get_fd(fd, EXTATTR_NAMESPACE_USER, name, value, size)
#define fsetxattr(fd, name, value, size, flags) \
	extattr_set_fd(fd, EXTATTR_NAMESPACE_USER, name, value, size)
#define fremovexattr(fd, name) \
	extattr_delete_fd(fd, EXTATTR_NAMESPACE_USER, name)
#else
#include <attr/xattr.h>
#endif

 /*
  * We minimize the overhead of checking taint bits, by predicting the most
  * likely branch: we assume that taint checks are disabled.
  * 
  * We minimize the overhead of propagating taint bits through low-level VM
  * instructions, by eliminating branches as much as possible.
  * 
  * The following is duplicated from zend_alloc.c. This should be global, and
  * renamed to something like ZEND_EXPECTED/UNEXPECTED, so that all code can
  * take advantage of this.
  */
#ifndef EXPECTED
#if (defined (__GNUC__) && __GNUC__ > 2 ) && !defined(__INTEL_COMPILER) && !defined(DARWIN) && !defined(__hpux) && !defined(_AIX)
#define EXPECTED(condition)   __builtin_expect(condition, 1)
#define UNEXPECTED(condition) __builtin_expect(condition, 0)
#else					/* __GNUC__ etc. */
#define EXPECTED(condition)   (condition)
#define UNEXPECTED(condition) (condition)
#endif					/* __GNUC__ etc. */
#endif

// -----------------------------------------------------------------------

#define DEFAULT_DIR_POLICY_NAME "DefaultDirPolicy"
#define ATTRNAME_CONTAINER "user.php-ctr-policy"
#define ATTRNAME_DATA "user.php-data-policy"
#define POLICY_MAX_BYTES 999999999

#define Z_TAINT_POLICY(zval)    (zval).mp
#define Z_TAINT_POLICY_P(zval_p)  Z_TAINT_POLICY(*(zval_p))
#define Z_TAINT_POLICY_PP(zval_pp)  Z_TAINT_POLICY(**(zval_pp))

#define Z_POLICY Z_TAINT_POLICY
#define Z_POLICY_P Z_TAINT_POLICY_P
#define Z_POLICY_PP Z_TAINT_POLICY_PP

#define TAINT_POLICY_CLEAR(mp, cleanup) \
  if (cleanup && mp) {  \
    multipolicy_dtor (mp); \
  }                         \
  mp = NULL

#define Z_TAINT_POLICY_CLEAR(zval, cleanup) TAINT_POLICY_CLEAR(Z_TAINT_POLICY(zval), cleanup)

multipolicy *multipolicy_ctor (void);
void multipolicy_dtor (multipolicy *mp);
void multipolicy_add (multipolicy *mp, zval *policy, long offset, long nbytes);
int multipolicy_to_array (multipolicy *mp, zval **retval);
int multipolicy_from_array (zval *arr, multipolicy **retval);

#define MP_ADDREF(mp) (mp)->refcountx++
#define MP_DECREF(mp) (mp)->refcountx--
#define MP_REFCNT_P(mp) (mp)->refcountx

void taint_policy_set (zval *zv, multipolicy *policy);

void resin_attach(zval *zv, const char *policy TSRMLS_DC);
void resin_detach(zval *zv, const char *policy TSRMLS_DC);

//#define Z_TAINT_POLICY_SET_P(zval_p, p) Z_TAINT_POLICY_SET ((*zval_p), p)

// YIPAL: Clone the policy object when cloning an object (if it exists)
// Assumes that the policy object has a clone method.
void taint_policy_clone (multipolicy **dst_mp_p, multipolicy *src_mp, long offset, long nbytes TSRMLS_DC);
void taint_policy_merge (multipolicy **dst_mp_p, multipolicy *op1_mp, multipolicy *op2_mp, int nbytes1, int nbytes2 TSRMLS_DC);
int  taint_policy_check (multipolicy *mp, char *type, zval *dest TSRMLS_DC);
int  taint_policy_check_eval (multipolicy *mp TSRMLS_DC);
int  taint_policy_check_eval_fh (zend_file_handle *h TSRMLS_DC);
zval *taint_policy_echo_filter (char *type, zval *dest, zval *expr TSRMLS_DC);

char *get_xattr (int fd, char *buf);
zval *create_instance (char *classname TSRMLS_DC, int nparams, ...);
int check_ctr_policy (char *path, int use_parent, char *type, char *dst TSRMLS_DC);

multipolicy *unserialize_multipolicy (char *buf TSRMLS_DC);


#define POLICY_CLONE_COLLAPSE	(-1)
#define POLICY_CLONE_EXACT		(-2)

#define Z_TAINT_POLICY_CLONE_EXACT(dst, src) taint_policy_clone(&(Z_TAINT_POLICY_P(dst)), (Z_TAINT_POLICY_P(src)), POLICY_CLONE_EXACT, 0 TSRMLS_CC)
#define Z_TAINT_POLICY_CLONE_COLLAPSE(dst, src) taint_policy_clone(&(Z_TAINT_POLICY_P(dst)), (Z_TAINT_POLICY_P(src)), POLICY_CLONE_COLLAPSE, 0 TSRMLS_CC)
#define Z_TAINT_POLICY_CLONE_SUB(dst, src, offset, nbytes) taint_policy_clone(&(Z_TAINT_POLICY_P(dst)), (Z_TAINT_POLICY_P(src)), offset, nbytes TSRMLS_CC)
#define Z_TAINT_POLICY_MERGE(dst, src1, src2) taint_policy_merge(&(Z_TAINT_POLICY_P(dst)), (Z_TAINT_POLICY_P(src1)), (Z_TAINT_POLICY_P(src2)), -1, -1 TSRMLS_CC)
#define Z_TAINT_POLICY_MERGE_CONCAT(dst, src1, src2) taint_policy_merge(&(Z_TAINT_POLICY_P(dst)), (Z_TAINT_POLICY_P(src1)), (Z_TAINT_POLICY_P(src2)), Z_STRLEN_P(src1), Z_STRLEN_P(src2) TSRMLS_CC)
#define Z_TAINT_POLICY_CHECK(policy, type, dest) taint_policy_check(policy, type, dest TSRMLS_CC)

#endif
