/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2006 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Sara Golemon <pollita@php.net>                               |
  +----------------------------------------------------------------------+
*/

/* $Id: pdo_user.c,v 1.1 2006/05/03 23:42:24 pollita Exp $ */

#include "taint_sql_int.h"
#include "ext/standard/info.h"

ZEND_DECLARE_MODULE_GLOBALS(pdo_user)

extern pdo_driver_t pdo_user_driver;

/* *********************************
   * PtrMap linked list management *
   ********************************* */

struct _php_pdo_user_llist {
	php_pdo_user_llist *next, *prev;
	php_pdo_user_data *data;
};

int php_pdo_user_ptrmap_map(php_pdo_user_data *data TSRMLS_DC)
{
	php_pdo_user_llist *el = emalloc(sizeof(php_pdo_user_llist));

	el->next = PDO_USER_G(ptr_map);
	el->prev = NULL;
	el->data = data;

	if (PDO_USER_G(ptr_map)) {
		PDO_USER_G(ptr_map)->prev = el;
	}

	PDO_USER_G(ptr_map) = el;

	return SUCCESS;
}

int php_pdo_user_ptrmap_unmap(php_pdo_user_data *data TSRMLS_DC)
{
	php_pdo_user_llist *el = PDO_USER_G(ptr_map);

	if (!el) {
		/* Nothing to unmap, not normal codepath...  */
		return FAILURE;
	}

	while (el) {
		if (el->data == data) {
			if (el->prev) {
				el->prev->next = el->next;
			} else {
				PDO_USER_G(ptr_map) = el->next;
			}
			if (el->next) {
				el->next->prev = el->prev;
			}
			efree(el);
			return SUCCESS;
		}
		el = el->next;
	}

	return FAILURE;
}

void *php_pdo_user_ptrmap_locate(zval *object TSRMLS_DC)
{
	php_pdo_user_llist *el = PDO_USER_G(ptr_map);

	if (!el) {
		/* Nothing to find */
		return NULL;
	}

	while (el) {
		if (el->data->object == object) {
			return el->data;
		}
		el = el->next;
	}

	return NULL;
}

void php_pdo_user_ptrmap_destroy(TSRMLS_D)
{
	php_pdo_user_llist *el = PDO_USER_G(ptr_map);

	while (el) {
		php_pdo_user_llist *next = el->next;

		efree(el);
		el = next;
	}

	PDO_USER_G(ptr_map) = NULL;
}
