/*
   +----------------------------------------------------------------------+
   | Zend Engine                                                          |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 Zend Technologies Ltd. (http://www.zend.com)      |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the Zend license,     |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt.                                 |
   | If you did not receive a copy of the Zend license and are unable to  |
   | obtain it through the world-wide-web, please send a note to          |
   | license@zend.com so we can mail you a copy immediately.              |
   +----------------------------------------------------------------------+
   | Authors: Wietse Venema <wietse@porcupine.org>                        |
   +----------------------------------------------------------------------+
*/
#include "taint.h"
#include "taint_sql.h"
#include "zval_pool.h"

static int zval_to_multipolicy(zval *zv, multipolicy **mp_p) {
  *mp_p = 0;

  if (zv && (Z_TYPE_P(zv) != IS_NULL && 
             Z_TYPE_P(zv) != IS_ARRAY && 
             Z_TYPE_P(zv) != IS_OBJECT)) {
    php_error(E_ERROR,"policy must either be an object, array, or null (%d)", Z_TYPE_P(zv));
    return -1;
  }
  
  if (!zv || Z_TYPE_P(zv) == IS_NULL) {
    return 0;
    //taint_policy_set (zv, NULL);
    //RETURN_NULL();
    
  } else if (Z_TYPE_P(zv) == IS_OBJECT) {
    multipolicy *_mp = multipolicy_ctor();
    multipolicy_add (_mp, zv, 0, POLICY_MAX_BYTES);
    *mp_p = _mp;
    return 0;
    //Z_POLICY_P(zv) = mp;
    //RETURN_NULL();

  } else if (Z_TYPE_P(zv) == IS_ARRAY) {
    if (multipolicy_from_array (zv, mp_p)) {
      php_error(E_ERROR, "Error converting array to multipolicy");
      return -1;
    }
    return 0;
  } 
  assert (0);
}

/* {{{ proto void taint_set_policy(mixed variable, obj securityPolicy) 
   Sets the policy object on <obj> to
   <securityPolicy>. <securityPolicy> may be null.*/
ZEND_API PHP_FUNCTION(taint_set_policy)
{
  zval   *zv, *policy_arg;
  int     argc = ZEND_NUM_ARGS();
  
  if (argc != 2)
    WRONG_PARAM_COUNT;
  if (zend_parse_parameters(argc TSRMLS_CC, "zz", &zv, &policy_arg) == FAILURE)
    RETURN_NULL();

  if (zv->type == IS_NULL)
    php_error(E_ERROR,"cannot attach policy to NULL");
  if (zv->type == IS_OBJECT || zv->type == IS_ARRAY)
    php_error(E_ERROR,"arrays, objects, and resources may not have policies");

  dprintf ("taint_set_policy, obj %p, policy %p, policy refcount %d\n", 
           zv, policy_arg, policy_arg->refcount);

  if (Z_TYPE_P(policy_arg) == IS_NULL) 
    Z_POLICY_P(zv) = NULL;
  else if (zval_to_multipolicy (policy_arg, &Z_POLICY_P(zv)))
    php_error (E_ERROR, "could not convert arg 2 to multipolicy");

  RETURN_NULL();
}

/* }}} */

/* {{{ proto void taint_get_policy(mixed variable) 
   Returns the policy object on <obj>. */
ZEND_API PHP_FUNCTION(taint_get_policy)
{
  zval   *zv;
  int     argc = ZEND_NUM_ARGS();
  
  if (argc != 1)
    WRONG_PARAM_COUNT;
  if (zend_parse_parameters(argc TSRMLS_CC, "z", &zv) == FAILURE)
    RETURN_NULL();

  if (zv->type == IS_NULL)
    RETURN_NULL();

  if (Z_POLICY_P(zv) == NULL)
    RETURN_NULL();

  zval *arr;
  if (multipolicy_to_array (Z_POLICY_P(zv), &arr)) {
    php_error(E_ERROR, "Error converting multipolicy to array");
    assert (0);
    RETURN_NULL();
  }

  RETURN_ZVAL (arr, 0, 1);
}
/* }}} */

#define PHP_STREAM_TO_ZVAL(stream, arg)					\
    php_stream_from_zval_no_verify(stream, arg);		\
    if (stream == NULL) {								\
        RETURN_NULL();									\
    }

/* {{{ proto void taint_set_file_policy(obj file, obj securityPolicy) 
   Sets the container policy object on <file> to
   <securityPolicy>. <securityPolicy> may be null.*/
ZEND_API PHP_FUNCTION(taint_set_file_policy)
{
	zval *zf, *policy_arg;
	int argc = ZEND_NUM_ARGS();
	php_stream *stream;
        multipolicy *mp = NULL;
  
	if (argc != 2)
		WRONG_PARAM_COUNT;
	if (zend_parse_parameters(argc TSRMLS_CC, "rz", &zf, &policy_arg) == FAILURE)
		RETURN_NULL();
	if (Z_TYPE_P(zf) != IS_RESOURCE)
		php_error(E_ERROR, "cannot set container policy on non-files");
	if (Z_TYPE_P(policy_arg) != IS_NULL && 
            Z_TYPE_P(policy_arg) != IS_OBJECT && 
            Z_TYPE_P(policy_arg) != IS_ARRAY)
		php_error(E_ERROR, "policy must be an object, array, or null");

        if (zval_to_multipolicy (policy_arg, &mp)) {
          php_error(E_ERROR, "could not convert zval to mp");
          RETURN_NULL();
        }
          
	PHP_STREAM_TO_ZVAL(stream, &zf);
	php_stream_set_policy(stream, 0, mp);
        multipolicy_dtor (mp);
	RETURN_NULL();
}

/* }}} */

/* {{{ proto void taint_get_file_policy(obj file)
   Returns the policy object on <file>. */
ZEND_API PHP_FUNCTION(taint_get_file_policy)
{
	zval *zf, *zmp = NULL;
        multipolicy *mp = 0;
	int argc = ZEND_NUM_ARGS();
	php_stream *stream;
  
	if (argc != 1)
		WRONG_PARAM_COUNT;
	if (zend_parse_parameters(argc TSRMLS_CC, "r", &zf) == FAILURE)
		RETURN_NULL();

	if (zf->type != IS_RESOURCE)
		php_error(E_ERROR, "argument isn't a file");

	PHP_STREAM_TO_ZVAL(stream, &zf);
	php_stream_get_policy(stream, 0, &mp);
        if (!mp)
          RETURN_NULL();

        multipolicy_to_array (mp, &zmp);
        multipolicy_dtor (mp);
        if (!zmp)
          RETURN_NULL();
        RETURN_ZVAL(zmp, 0, 0);
}
/* }}} */


ZEND_API PHP_FUNCTION(taint_foo)
{
  zval *zv;
  int argc = ZEND_NUM_ARGS();

  if (argc != 1)
    WRONG_PARAM_COUNT;
  if (zend_parse_parameters (argc TSRMLS_CC, "z", &zv) == FAILURE)
    RETURN_NULL();

  dprintf ("zval_type: %d (%p)\n", EG(uninitialized_zval_ptr)->type, EG(uninitialized_zval_ptr));

  /*
  php_printf ("FOO: zval %p\n", zv);
  php_printf ("  type %d\n", zv->type);
  php_printf ("  policy %p\n", zv->policy);
  if (zv->policy) 
  php_printf ("  policy refcount %u\n", zv->policy->refcount);
  
  if (zv->type == IS_OBJECT) {
    php_printf ("  object id %u, handlers %p\n", 
                zv->value.obj.handle, zv->value.obj.handlers);
  }
  */

  RETURN_ZVAL(zv, 1, 0);
}

/* ---------------------------------------------------- */
/* SQL Parser code */

static void *pdo_user_malloc_wrapper(size_t x) { return emalloc(x); }
static void pdo_user_free_wrapper(void *x) { efree(x); }

/* {{{ taint_parse_sql (string sql)
Compile a SQL statement into a parse tree */
ZEND_API PHP_FUNCTION(taint_parse_sql)
{
	char *sql;
        zval *sql_zv;
	int sql_len;
	php_pdo_user_sql_tokenizer T;
	php_pdo_user_sql_token token;
	void *pParser;
        parse_return_data pReturn;
        pReturn.return_val = return_value;
        pReturn.parse_error = 0;
        INIT_PZVAL (return_value);

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &sql, &sql_len) == FAILURE) {
          RETURN_NULL();
	}

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &sql_zv) == FAILURE) {
          RETURN_NULL();
	}

	T.start = sql;
	T.end = sql + sql_len;
	pParser = php_pdo_user_sql_parserAlloc( pdo_user_malloc_wrapper );

	while (PU_END != php_pdo_user_sql_get_token(sql_zv, sql, &T, &token TSRMLS_CC)) {
          if (token.id != PU_WHITESPACE) {
            //php_printf ("Parsing token %d (%d bytes) [%s]\n", token.id, token.token_len, token.token);
            php_pdo_user_sql_parser(pParser, token.id, token, &pReturn);
            
            if (pReturn.parse_error) {
              zval_dtor(pReturn.return_val); // clean up the partial parse tree
              php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure parsing SQL statement at: (%d) %s", token.id, token.token);
              php_pdo_user_sql_parserFree(pParser, pdo_user_free_wrapper);
              RETURN_NULL();
            }
          } else if (token.mp) {
            multipolicy_dtor (token.mp);
            token.mp = NULL;
          }
	}
	php_pdo_user_sql_parser(pParser, 0, token, &pReturn);
	php_pdo_user_sql_parserFree(pParser, pdo_user_free_wrapper);

        if (token.mp) {
          multipolicy_dtor (token.mp);
          token.mp = NULL;
        }

        if (return_value->type == IS_NULL) {
          zval_dtor(pReturn.return_val); // clean up the partial parse tree
          php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure parsing SQL statement at: %s", token.token);
          RETURN_NULL();
        }
}
/* }}} */


/* ---------------------------------------------------- */
/* test */

PHP_FUNCTION(test_null_filter)
//zval *taint_policy_null_filter (zval *expr TSRMLS_DC) {
	// returns the zval to be printed, after processing through a
	// filter function "null_filter()".
{
	char *filter_fname = "null_filter";
	zval *expr;
	
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &expr) == FAILURE) { 
		php_error(E_ERROR, "wrong parameter");		
        RETURN_NULL(); 
	}

	//zval fn_name;
	//INIT_PZVAL(&fn_name);
	//ZVAL_STRING(&fn_name, filter_fname, 1);
	zval *fn_name_p = zval_pool_get(ZVAL_POOL_NULL_FILTER);
	//printf("test_null_filter: fn_name_p: %p\n", fn_name_p);
	
	zval type_name;
	INIT_PZVAL(&type_name);
	ZVAL_STRING(&type_name, "null", 1);
	zval *type_name_p = &type_name;
	
	zval **params[] = { &type_name_p, &expr };
	zval *retval = 0;
	
	int rc = call_user_function_ex(CG(function_table), NULL, fn_name_p, &retval,
								   2, params, 0, NULL TSRMLS_CC);
	//zval_dtor(&fn_name);
	zval_dtor(&type_name);
	//zval_dtor(&null_zval);
	
	if (rc == SUCCESS) {
	} else {
		zend_error (E_WARNING, "call to null_filter failed");
	}

	RETURN_NULL();
}

/* ---------------------------------------------------- */

/* Request that taint()/untaint() arguments are passed by reference. */

ZEND_BEGIN_ARG_INFO(taint_force_first_arg_by_reference, 0)
ZEND_ARG_PASS_INFO(1)
ZEND_END_ARG_INFO()

ZEND_BEGIN_ARG_INFO(taint_force_two_args_by_reference, 0)
ZEND_ARG_PASS_INFO(1)
ZEND_ARG_PASS_INFO(1)
ZEND_END_ARG_INFO()

/* Function table. */

static function_entry taint_functions[] = {
	PHP_FE(taint_set_policy, taint_force_two_args_by_reference)
	PHP_FE(taint_get_policy, taint_force_first_arg_by_reference)
	PHP_FE(taint_set_file_policy, taint_force_two_args_by_reference)
	PHP_FE(taint_get_file_policy, taint_force_first_arg_by_reference)
	PHP_FE(taint_foo, taint_force_first_arg_by_reference)
	PHP_FE(taint_parse_sql, taint_force_first_arg_by_reference)
	PHP_FE(test_null_filter, taint_force_first_arg_by_reference)
	0,
};

PHP_MINFO_FUNCTION(taint)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "taint support", "present");
	php_info_print_table_end();
}

zend_module_entry taint_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	TAINT_EXT_NAME,
	taint_functions,		/* functions */
	0,				/* minit */
	0,				/* mshutdown */
	0,				/* rinit */
	0,				/* rshutdown */
	PHP_MINFO(taint),		/* minfo */
#if ZEND_MODULE_API_NO >= 20010901
	TAINT_EXT_VERSION,
#endif
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_TAINT
ZEND_GET_MODULE(taint)
#endif
