<?php
/*
 *  include this file using "require" 
 */

define("POLICY_SUFFIX", "_policy");
define("POLICY_MAX_BYTES", 999999999);

$legal_fcall_names = array ('max', 'min', 'count', 'if', 'group_concat', 
                            'length', 'found_rows', 'sum', 'ifnull', 
                            'version', 'lower', 'coalesce', 'greatest', 
                            'concat', 'least', 'year', 'month',
                            'date_format', 'values');
$null_policy_fcalls = $legal_fcall_names;
$null_policy_labels = array ('current_timestamp');

$stderr = fopen ("php://stderr", 'w');

function perr ($s) {
  global $stderr;
  fwrite ($stderr, $s);
  fflush($stderr);
}

function print_r_stderr ($x) {
  foreach ($x as $k => $v) {
    perr ("$k => $v\n");
  }
}

function debug_echo ($s) {
  if (false)
    perr ($s);
}

function mp_to_str ($mp) {
  $ret = array ();
  if ($mp) {
    foreach ($mp as $ent) {
      if ($ent['nbytes'] >= POLICY_MAX_BYTES) {
        $len = "inf";
      } else {
        $len = $ent['nbytes'];
      }
      $ret[] = sprintf ("$len@%d: %s", $ent['offset'], $ent['policy']);
    }
  }
  return implode (', ', $ret);
}

function mp_collapse ($mp) {
  if ($mp && sizeof($mp) > 0) {
    $ret = array();
    foreach ($mp as $ent) {
      $ret[] = array ('offset' => 0, 
                      'nbytes' => POLICY_MAX_BYTES,
                      'policy' => $ent['policy']);
    }
    return $ret;
  }
  return NULL;
}

function mp_merge($mp1, $mp2, $n1, $n2) {
  // PHP version slower, but easily accessible from PHP.
  
  // $mp1 multipolicy1
  // $mp2 multipolicy2
  // $n1 string length from p1
  // $n2 string length from p2
  //
  // If lengths n1 and n2 are non-negative, 
  // Outputs a MultiPolicy containing all the policies from p1 and p2 that are within [n1, n1+n2]
  
  $new_mp = array ();
  
  $collapse = ($n1 < 0) || ($n2 < 0);
  $i = 0;
  
  if ($mp1) {
    foreach ($mp1 as $ent) {
      $new_mp[] = array ('offset' => $collapse ? 0 : $ent['offset'],
                         'nbytes' => $collapse ? POLICY_MAX_BYTES : (($ent['offset'] + $ent['nbytes'] > $n1) ? 
                                                                     ($n1 - $ent['offset']) : 
                                                                     $ent['nbytes']),
                         'policy' => $ent['policy']);
      
    }
  }
  if ($mp2) {
    foreach ($mp2 as $ent) {
      $new_mp[] = array ('offset' => $collapse ? 0 : $ent['offset'] + $n1,
                         'nbytes' => $collapse ? POLICY_MAX_BYTES : (($ent['offset'] + $ent['nbytes'] > $n1) ? 
                                                                     ($n1 - $ent['offset']) : 
                                                                     $ent['nbytes']),
                         'policy' => $ent['policy']);
    }
  }

  switch (sizeof($new_mp)) {
  case 0:
    return null;
  case 1:
    // pass
    break;
  case 2:
    // same policy
    if ($new_mp[0]['policy'] == $new_mp[1]['policy']) {
      if ($new_mp[0]['offset'] > $new_mp[1]['offset']) 
        $new_mp = array ($new_mp[1], $new_mp[0]);

      // overlap
      if ($new_mp[0]['offset'] + $new_mp[0]['nbytes'] >= $new_mp[1]['offset']) {
        $nbytes = max($new_mp[0]['offset'] + $new_mp[0]['nbytes'],
                      $new_mp[1]['offset'] + $new_mp[1]['nbytes']) - $new_mp[0]['offset'];
        $key = sprintf ("%d:%d:%d", 0, $new_mp[0]['offset'], $nbytes);
        $new_mp = array($key => $new_mp[0]['policy']);
      }
    }
    break;
  default:
    // mulitple kinds of polices
    // we may use a map (keyed on policy) to merge them
    // but it might be too slow
    break;
  }
  return $new_mp;
}

function __policy_has_annotation($mp, $classname) {
  // $mp is the hashtable representation of a multipolicy
  if ($mp) {
    foreach ($mp as $ent)
      if ($ent['policy'] instanceof $classname) {
        return true;
      }
  }
  return false;
}


class DummyPolicy {
	public $s;
	public function __construct($n=NULL) { $this->s = $n; }
	public function __toString() { return $this->s; }
	public function __output($type, $dst) {}
}


function exprlist_to_string ($node) {
  $a = array ();
  foreach ($node as $val) {
    $a[] = expr_to_string ($val);
  }
  return implode (', ', $a);
}

function insert_to_string ($node) {
  $table = $node['table'];
  $fields = implode (', ', $node['fields']);

  $insertgroups = array ();
  foreach ($node['data'] as $insertgroup) {
    $insertgroups[] = '(' . exprlist_to_string ($insertgroup). ')';
  }
  $insertgroups = implode (', ', $insertgroups);
  $ondup = '';
  if (array_key_exists ('ondup', $node)) {
    print_r($node['ondup']);
    $ondup = ' ON DUPLICATE KEY UPDATE ' . set_to_string ($node['ondup']);
  }
  return "INSERT INTO $table ($fields) VALUES $insertgroups$ondup" ;
}

function insert_select_to_string ($node) {
  $table = $node['table'];
  $fields = implode (', ', $node['fields']);
  $query = select_to_string ($node['query']);
  $ondup = '';
  if (array_key_exists ('ondup', $node))
    $ondup = ' ON DUPLICATE KEY UPDATE ' . set_to_string ($node['ondup']);
  return "INSERT INTO $table ($fields) $query$ondup";
}

function select_to_string ($node) {

  $opts = implode(', ', $node['opts']);
  if ($opts != '')
    $opts .= ' ';

  $fields = exprlist_to_string ($node['fields']);

  $from = '';
  if ($node['from'] != NULL)
    $from = ' FROM ' . fromlist_to_string ($node['from']);

  $modifiers = NULL;

  if (is_array ($node['modifiers'])) {
    $modifiers = modifiers_to_string ($node['modifiers']);
  }

  $ret = "SELECT $opts$fields$from";
  if ($modifiers) 
    $ret .= " $modifiers";
  return $ret;
}

function from_to_string ($node){
  switch ($node['type']) {
  case 'alias':
    return expr_to_string ($node);
  case 'table':
    return expr_to_string ($node['name']);
  case 'join':
    $s = from_to_string ($node['tableref1']);
    $s .= ' ' . strtoupper ($node['join-type']) . ' ';
    $s .= from_to_string ($node['tablefact2']);
    if ($node['condition']) {
      switch ($node['condition']['type']) {
      case 'on':
        $s .= ' ON ' . expr_to_string ($node['condition']['val']);
        break;
      case 'using':
        $s .= ' USING (' . exprlist_to_string ($node['condition']['val']) . ')';
        break;
      }
    }
    return $s;
  case 'parengroup':
    return '('. fromlist_to_string ($node['parengroup']). ')';
  default:
    perr ("Unexpected FROM type: " . $node['type'] . "\n");
    //print_r ($node);
    return;
  }
}

function fromlist_to_string ($node) {
  $list = array();
  foreach ($node as $e) {
    $list[] = from_to_string($e);
  }
  return implode(', ', $list);
}

function join_to_string ($node) {
  return from_to_string ($node['table1']) . ' ' . strtoupper ($node['join-type']) . " JOIN " . expr_to_string ($node['table2']) . " ON " . expr_to_string ($node['on']);
}

function where_to_string ($node) {
  return 'WHERE ' . expr_to_string ($node);
}
function orderby_to_string ($node) {
  $list = array ();
  foreach ($node as $orderelement) {
    $list[] = expr_to_string ($orderelement['by']) . ($orderelement['direction'] ? ' ' . strtoupper ($orderelement['direction']) : '');
  }
  return 'ORDER BY ' . implode (', ', $list);
}
function modifiers_to_string ($node) {
  $out = array();
  if (array_key_exists ('where', $node)) {
    $out[] = where_to_string ($node['where']);
  }

  if (array_key_exists ('group-by', $node)) {
    $out[] = 'GROUP BY ' . exprlist_to_string ($node['group-by']);
  }
  if (array_key_exists ('order-by', $node)) {
    $out[] = orderby_to_string ($node['order-by']);
  }
  if (array_key_exists ('limit', $node)) {
    $lnode = $node['limit'];
    $lim = 'LIMIT ';
    if (array_key_exists ('from', $lnode)) {
	  $lim = $lim . $lnode['from']. ', ';
    }
    $lim = $lim . $lnode['to'];
    $out[] = $lim;
  }
  return implode (' ', $out);
}

function update_to_string ($node) {
  $s = 'UPDATE ' . $node['table'] . ' SET ' . set_to_string ($node['set']);
  if ($node['where']) {
    $s .= ' ' . where_to_string ($node['where']);
  }
  return $s;
}

function delete_to_string ($node) {
  // This only handles single table syntax
  $s = 'DELETE FROM ' . $node['table'];
  if ($node['where']) {
    $s .= ' ' . where_to_string ($node['where']);
  }
  return $s;
}

function set_to_string ($node) {
  $s = array ();
  foreach ($node as $varname => $expr) {
    $s[] = "$varname=" . expr_to_string ($expr);
  }
  return implode (', ', $s);
}

function expr_to_string ($node) {
  global $legal_fcall_names;

  if (is_array ($node)) {
    $type = $node['type'];

    if ($type == 'condition') {
      return "(" . expr_to_string ($node['op1']) . ' ' . strtoupper($node['condition']) . ' ' . expr_to_string ($node['op2']) . ")";
    } else if ($type == 'math') {
      return "(" . expr_to_string ($node['op1']) . ' ' . strtoupper($node['operation']) . ' ' . expr_to_string ($node['op2']) . ")";
    } else if ($type == 'math-unary') {
      return "(" . strtoupper($node['operation']) . ' ' . expr_to_string ($node['op1']) . ")";
    } else if ($type == 'statement') {
      return statement_to_string ($node);

    } else if (array_key_exists ('table', $node) && array_key_exists ('field', $node)) {
      return $node['table'] . '.' . $node['field'];

    } else if ($type == 'label') {
      $s = $node['val'] . '';
      if (array_key_exists ('table', $node))
        $s = $node['table'] . '.' . $s;
      if (array_key_exists ('database', $node))
        $s = $node['database'] . '.' . $s;
      return $s;

    } else if ($type == 'num') {
      return $node['val'] . '';

    } else if ($type == 'string') {
      return $node['val'] . '';
      //return $node['val'] . "(policy: " . taint_get_policy($node['val']) . ")";

    } else if ($type == 'fcall') {
      if (in_array (strtolower ($node['fname']), $legal_fcall_names)) {
        return $node['fname'] . "(" . exprlist_to_string ($node['args']) . ")";
      } else {
        perr ("Unexpected or illegal function name: " . $node['fname'] . "\n");
        return;
      }

    } else if ($type == 'alias') {
      if (array_key_exists ('field', $node))
        $f = $node['field'];
      else if (array_key_exists ('table', $node))
        $f = $node['table'];
      else if (array_key_exists ('orig', $node))
        $f = $node['orig'];
      else {
        perr ("Unexpected alias structure\n");
        print_r ($node);
        return;
      }
      $e = expr_to_string ($f);
      if (array_key_exists ('paren', $node))
        $e = "($e)";
      return $e . (($node['as']) ? (' AS ' . $node['as']) : '');

    } else if ($type == 'null') {
      return "NULL";

    } else if ($type == 'inexpr-stmt') {
      return "(" . expr_to_string($node['inexpr']) . ")";

    } else if ($type == 'inexpr-grouplist') {
      $list = array();
      foreach ($node['inexpr'] as $e) {
        $list[] = expr_to_string($e);
      }
      return "(" . implode(', ', $list) . ")";
      
    } else if ($type == 'distinct') {
      return "DISTINCT ". expr_to_string($node['distinct']);
      
    } else {
      perr ("Unexpected expr type1: " . $type . "\n");
      perr (print_r ($node,true));
      return;
    }
  } else {
    perr ("Expr not an array: " . $node . "\n");
    //debug_print_backtrace ();
    return;
  }
}

function statement_to_string ($node) {
  switch ($node['statement']) {
  case 'insert':
    return insert_to_string ($node);
  case 'insert-select':
    return insert_select_to_string ($node);
  case 'select':
    return select_to_string ($node);
  case 'update':
    return update_to_string ($node);
  case 'delete':
    return delete_to_string ($node);
  case 'create-temp':
    return createtemp_to_string ($node);
  default:
    perr ("Unexpected statement type: " . $node['statement'] . "\n");
    return '';
  }
}

function notranslate_to_string ($node) {
  return implode (' ', $node['values']);
}

function createtemp_to_string ($node) {
  return "CREATE TEMPORARY TABLE " . $node['table'] . ' ' . statement_to_string ($node['query']);
}

function parse_tree_to_string ($node) {
  switch ($node['type']) {
  case 'statement':
    return statement_to_string ($node);
    break;
  case 'notranslate':
    return notranslate_to_string ($node);
    break;
  default:
    perr ("Unexpected node type: " . $node['type'] . "\n");
  }
}

function get_func_cols ($select_node) {
  // returns ids function columns
  $func_cols = array ();
  foreach ($select_node['fields'] as $k => $field) {
    $f = $field;
    if ($field['type'] == 'alias')
      $f = $field['orig'];

    if ($f['type'] == 'fcall')
      $func_cols[] = ($field['type'] == 'alias' ? $field['as'] : $f['val']);
  }
  return $func_cols;
}

function is_func_from ($from_node, $col_label) {
  
  if ($from_node['type'] == 'join') {
    if (is_func_from ($from_node['tableref1'], $col_label))
      return true;
    if (is_func_from ($from_node['tablefact2'], $col_label))
      return true;

  } else if ($from_node['type'] == 'alias') {

    if ($from_node['orig']['type'] == 'label') {
      if ($from_node['as'] && $from_node['as'] == $col_label['table']) {
        // it's a regular table without function calls
        return false;
      }
      if (array_key_exists('table', $col_label) && $from_node['orig']['val'] == $col_label['table']) {
        // it's a regular table without function calls
        return false;
      }
      if (array_key_exists ('table', $col_label)) {
        // no table specified, so it must be a column from this table.
        return false;
      }

    } else if ($from_node['orig']['type'] == 'statement' &&
               $from_node['orig']['statement'] == 'select') {
      if (array_key_exists ('table', $col_label) && $from_node['as'] != $col_label['table']) {
        // table name doesn't match (subselect must use AS)
        return false;
      }
      
      $funcs = get_func_cols ($from_node['orig']);
      //echo "FUNCS: " . print_r ($funcs, true); 
      return (in_array ($col_label['val'], $funcs));
    }
  }
  return false;
}

function is_func ($select_node, $col_label) {
  if (is_array ($select_node) && array_key_exists ('statement', $select_node) && $select_node['statement'] == "select") {
    //echo "CHECKING: " . $col_label['table'] . '.' . $col_label['val'] . " IN " . print_r ($select_node['from'], true);
    foreach ($select_node['from'] as $k => $fromclause) {
      if (is_func_from ($fromclause, $col_label))
        return true;
    }
  }
  return false;
}

function add_policies_select (&$tree, $add_placeholders=false) {
  if (is_array ($tree) && array_key_exists ('statement', $tree) && $tree['statement'] == "select") {

    // recurse to subselects that will show up in the results
    //perr ("FOO: " . print_r ($tree['from'], true) . "\n");
    foreach ($tree['from'] as $k => $fromclause) {
      if ($fromclause['type'] == 'join' && 
          $fromclause['tablefact2']['orig']['type'] == 'statement' &&
          $fromclause['tablefact2']['orig']['statement'] == 'select') {
        add_policies_select ($tree['from'][$k]['tablefact2']['orig']);
      }
    }

    // add columns
    $newfields = $tree['fields'];

    foreach ($tree['fields'] as $f) {
      if ($f['type'] == 'label' && $f['val'] != '*' && !is_func ($tree, $f)) {
        $a = $f;
        $a['val'] = $f['val'] . POLICY_SUFFIX;
        $newfields[] = $a;

      } else if ($f['type'] == 'alias' && $f['orig']['type'] == 'label' && !is_func ($tree, $f['orig'])) {
        $a = $f;
        $a['orig']['val'] = $f['orig']['val'] . POLICY_SUFFIX;
        $a['as'] = (($f['as']) ? ($f['as'] . POLICY_SUFFIX) : '');
        $newfields[] = $a;
      } else if ($add_placeholders) {
        $newfields[] = array ('type' => 'string', 'val' => "''");
      }
    }
    $tree['fields'] = $newfields;
  }
  return $tree;
}

function get_expr_policy ($expr, $dst_col=NULL) {
  switch ($expr['type']) {
  case 'string':
  case 'num':
    //stats_begin('get_policy');
    $mp = taint_get_policy ($expr['val']);
    //stats_end('get_policy');
    return mp_collapse ($mp);
  case 'math':
    $mp1 = get_expr_policy ($expr['op1'], $dst_col);
    $mp2 = get_expr_policy ($expr['op2'], $dst_col);
    return mp_merge ($mp1, $mp2, -1, -1);
  case 'math-unary':
    return get_expr_policy ($expr['op1'], $dst_col);
  case 'null':
    return NULL;
  case 'label':
    global $null_policy_labels;
    if (in_array ($expr['val'], $null_policy_labels))
      return NULL;

    if ($expr['val'] == $dst_col)
      return NULL;  // YIPAL: this is broken, what we should do is
                    // remove the col_policy='' entry if all other
                    // operands have empty policy objects because then
                    // we'll know that the policy will stay the same.
                    // Ultimately, we could replace this with a
                    // SELECT.
    perr ("Unexpected1 expr label for dst ($dst_col) has input (" . $expr['val'] . ")\n");
    return;
  case 'fcall':
    if (in_array ($expr['fname'], $null_policy_fcalls))
      return NULL;
    perr ("Unexpected2 expr label for dst ($dst_col) has input (" . $expr['fname'] . ")\n");
    return;
    
  default:
    perr ("Unexpected expr type2: " . $expr['type'] . "\n");
    return;
  }
}

function field_add_suffix($f) {
  if (substr($f, 0, 1) == '`') {
    return substr($f, 0, strlen($f) - 1) . POLICY_SUFFIX . '`';
  } else {
    return $f . POLICY_SUFFIX;
  }
}

function add_policies_insert (&$tree) {
  // add columns
  //stats_begin ('append-old');
  $newfields = $tree['fields'];
  foreach ($tree['fields'] as $f) {
    $newfields[] = field_add_suffix($f);
  }
  $tree['fields'] = $newfields;
  //stats_end ('append-old');

  if ($tree['statement'] == 'insert') {
    // add policies for insertgroups
    $oldinsertgroups = $tree['data'];
    $newinsertgroups = array ();
    foreach ($oldinsertgroups as $oldinsertgroup) {
      $newinsertgroup = $oldinsertgroup;
      foreach ($oldinsertgroup as $item) {
        //stats_begin('get_expr_policy');
        $item_policy = get_expr_policy ($item, NULL);
        //stats_end('get_expr_policy');

        //stats_begin ('serialize');
        $item_policy = $item_policy ? "'" . mysql_real_escape_string(serialize($item_policy)) . "'" : 'NULL';
        //$item_policy = serialize($item_policy);
        //stats_end ('serialize');
        
        //stats_begin ('escape');
        //$item_policy = ($item_policy);
        //stats_end ('escape');
        $newinsertgroup[] = array ('type' =>'string', 'val' => $item_policy);
      }
      $newinsertgroups[] = $newinsertgroup;
    }
    $tree['data'] = $newinsertgroups;
  } else if ($tree['statement'] == 'insert-select') {
    // This assumes that the schema in the DB puts all the
    // policies after the data columns.  This is true in
    // hotcrp/schema.sql.
    $tree['query'] = add_policies_select ($tree['query'], true);
  } else {
    perr ("Unexpected insert type: " . $tree['statement'] . "\n");
    return NULL;
  }
  return $tree;
}

function add_policies_update (&$tree) {
  if (is_array ($tree)) {
    if (array_key_exists ('statement', $tree) && 
        ($tree['statement'] == "update")) {

      $newset = $tree['set'];
      foreach ($tree['set'] as $colname => $colval) {

        if ($colval['type'] == 'string' || $colval['type'] == 'num' ||
            $colval['type'] == 'math' || $colval['type'] == 'math-unary') {
          $policy = get_expr_policy ($colval, $colname);
          $policy = serialize($policy);
          $policy = mysql_real_escape_string ($policy);
          $newcolval = array ('type' => 'string', 'val' => '\'' . $policy . '\'');
          $newset[$colname . POLICY_SUFFIX] = $newcolval;
        } else {
          // Currently, this doesn't support UPDATE with a subselect, but it could.
          perr ("Unexpected col type for update: " . $colval['type'] . "\n");
        }
      }
      $tree['set'] = $newset;
    }
  }
  return $tree;
}

function add_policies_createtemp (&$tree) {
  if (is_array ($tree))
    if (array_key_exists ('statement', $tree) && 
        ($tree['statement'] == 'create-temp'))
      add_policies_select ($tree['query']);
  return $tree;
}

$stats = array ();
$active = array ();
function stats_begin ($key) {
  global $active;
  $active[$key] = gettimeofday(true);
}
function stats_end ($key) {
  global $active, $stats;
  if (!array_key_exists ($key, $stats))
    $stats[$key] = array(0, 0);
  $stats[$key][0] += (gettimeofday(true) - $active[$key]);
  $stats[$key][1]++;
}

function stats_report () {
  global $stats;
  foreach ($stats as $k => $dat) {
    $usec = 1000000 * ($dat[0] / $dat[1]);
    perr (sprintf ("%-15s %7d %10.3f\n", 
                   $k, $dat[1], $usec));
  }
}
function stats_reset () {
  global $stats, $active;
  $stats = array ();
  $active = array ();
}

function sql_translate ($q, $debug=false) {
  $bypass = array ('lock', 'unlock', 'drop', 'show', 'set');
  foreach ($bypass as $cmd)
    if (!substr_compare ($q, $cmd, 0, strlen($cmd), true))
      return $q;

  $tree = taint_parse_sql ($q);
  if ($tree == NULL) { 
    perr ("PARSE ERROR ON: $q\n");
    return NULL;
  }

  if ($debug) {
    print_r ($tree);
  }

  switch ($tree['statement']) {
  case 'select':
    $tree = add_policies_select ($tree);
    break;
  case 'insert':
  case 'insert-select':
    $tree = add_policies_insert ($tree);
    break;
  case 'update':
    $tree = add_policies_update ($tree);
    break;
  case 'create-temp':
    $tree = add_policies_createtemp ($tree);
    break;
  case 'delete':
    break;
  default:
    perr("unexpected statement " . $tree['statement'] . "\n");
  }
  if ($debug) 
    print_r ($tree);

  return statement_to_string ($tree);
}

function sql_translate_express ($q) {
  //stats_begin('translate-total');

  //stats_begin('parse');
  if (!($tree = taint_parse_sql ($q))) { 
    perr ("PARSE ERROR ON: $q\n");
    return NULL;
  }
  //stats_end('parse');

  switch ($tree['statement']) {
  case 'select':
    //stats_begin('select');
    $tree = add_policies_select ($tree);
    //stats_end('select');
    break;
  case 'insert':
  case 'insert-select':
    //stats_begin('insert');
    $tree = add_policies_insert ($tree);
    //stats_end('insert');
    break;
  case 'update':
    //stats_begin('update');
    $tree = add_policies_update ($tree);
    //stats_end('update');
    break;
  case 'create-temp':
    //stats_begin('createtemp');
    $tree = add_policies_createtemp ($tree);
    //stats_end('createtemp');
    break;
  case 'delete':
    break;
  default:
    perr("unexpected statement " . $tree['statement'] . "\n");
  }

  //stats_begin('tostring');
  $s = statement_to_string ($tree);
  //stats_end('tostring');

  //stats_end('translate-total');
  return $s;
}

function resin_mysql_fetch ($result, $type) {
  /* If you use a query where some return columns are named by number
     for example, "SELECT c1, 0 FROM t1", the second field "0" will
     clobber the first field in results from resin_mysql_fetch_row.
     Always provide an alias for number columns.
  */

  //stats_begin("mysql_fetch");
  if (!$result || !($dat = __resinorig_mysql_fetch_array ($result)))
    return false;
  //stats_end("mysql_fetch");

  //stats_begin("postproc");
  $ret_arr = array ();
  $ret_arr_assoc = array ();
  $ret_row = array ();
  $ret_obj = (object) array ();

  $key_to_idx = array ();
  $last_idx = -1;

  foreach ($dat as $k => $v) {
    if (strlen($k) > strlen(POLICY_SUFFIX) && 
        !substr_compare ($k, POLICY_SUFFIX, strlen ($k) - strlen(POLICY_SUFFIX), strlen(POLICY_SUFFIX))) {
      $policy_idx = $last_idx;
      unset ($ret_arr[$policy_idx]);
      unset ($ret_row[$policy_idx]);

      $dat_key = substr($k, 0, strlen ($k) - strlen(POLICY_SUFFIX));
      $dat_idx = $key_to_idx[$dat_key];
      
      if (!$v || !strlen($v))
        $policy = null;
      else {
        //stats_begin("unserialize");
        $policy = unserialize($v);
        //stats_end("unserialize");
      }

      //$policy = (!$v || strlen ($v) == 0) ? null : unserialize ($v);

      if ($policy === false) {
        perr ("Could not unserialize policy ($v)\n");
      } else if ($policy) {
        //if ($policy) perr ("setting on ($dat_key, $dat_idx) policy to: ($policy) ($v)\n");
        //debug_zval_dump($policy);

        if ($ret_arr[$dat_key])
          taint_set_policy ($ret_arr[$dat_key], $policy);
        if ($ret_arr[$dat_idx])
          taint_set_policy ($ret_arr[$dat_idx], $policy);
        if ($ret_arr_assoc[$dat_key])
          taint_set_policy ($ret_arr_assoc[$dat_key], $policy);
        if ($ret_row[$dat_idx])
          taint_set_policy ($ret_row[$dat_idx], $policy);
        if ($ret_obj->{$dat_key}) 
          taint_set_policy ($ret_obj->{$dat_key}, $policy);
      }
    } else {
      $ret_arr[$k] = $v;
      if (is_numeric ($k)) {
        $last_idx = $k;
        $ret_row[$k] = $v;
      } else {
        $key_to_idx[$k] = $last_idx;
        $ret_obj->{$k} = $v;
        $ret_arr_assoc[$k] = $v;
      }
    }
  }  
  /*
  perr (print_r ($dat, true));
  perr (print_r ($ret_arr, true));
  perr (print_r ($ret_row, true));
  perr (print_r ($ret_obj, true));
  */
  //stats_end("postproc");

  switch ($type) {
  case 'array':
    return $ret_arr;
  case 'assoc':
    return $ret_arr_assoc;
  case 'row':
    return $ret_row;
  case 'object':
    return $ret_obj;
  default:
    perr ("resin_mysql_fetch: unexpected type: $type\n");
    return false;
  }
}

function mysql_fetch_row    ($result) { return resin_mysql_fetch ($result, 'row'); }
function mysql_fetch_array  ($result) { return resin_mysql_fetch ($result, 'array'); }
function mysql_fetch_object ($result) { return resin_mysql_fetch ($result, 'object'); }
function mysql_fetch_assoc  ($result) { return resin_mysql_fetch ($result, 'assoc'); }

function str_ends_with($s, $ending) {
  $len = strlen($ending);
  $send = substr($s, strlen($s) - $len);
  return $send == $ending;
}

function mysql_fetch_field ($result) {
  while (true) {
    $f = __resinorig_mysql_fetch_field($result);
    if (!$f || !str_ends_with($f->name, POLICY_SUFFIX)) {
      return $f;
    }
  }
}

function mysql_num_fields ($result) {
  $pos = mysql_field_tell($result);

  $nf = __resinorig_mysql_num_fields($result);
  $ret = 0;
  for ($i = 0; $i < $nf; $i++) {
    $f = __resinorig_mysql_fetch_field($result, $i);
    if (!str_ends_with($f->name, POLICY_SUFFIX)) {
      $ret++;
    }
  }

  mysql_field_seek($result, $pos);
  return $ret;
}

function mysql_query ($query, $link, $debug=false) {
  if ($debug) 
    perr ("QUERY: " . $query . "\n");
  $q = sql_translate ($query);
  if ($debug)
    perr ("TRANS: " . $q . "\n");
  return __resinorig_mysql_query ($q, $link);
}

// XSS
function resin_attach(&$obj, $p) {
	if (!$p)
		return $obj;
	$parr = taint_get_policy($obj);
	$policy = $parr;	/* just to be sure we copied-by-value */
	if (!$policy) {
		$policy = array();
	}
	$policy[] = array ('offset' => 0, 
                      	   'nbytes' => POLICY_MAX_BYTES,
                      	   'policy' => $p);
	taint_set_policy($obj, $policy);
	return $obj;
}

function resin_detach(&$obj, $p) {
	$parr = taint_get_policy($obj);
	$policy = $parr;	/* just to be sure we copied-by-value */
	if (!$policy) {
		return;
	}

	$filter = create_function('$e', 'return !($e["policy"] instanceof ' . $p . ');');
	$policy = array_filter($policy, $filter);
	taint_set_policy($obj, $policy);
	return $obj;
}

// cross-site/external-source policy
// e.g., SQL injection, cross-site scripting
class XSPolicy extends DummyPolicy {
	const XS = 'XS';

	function __construct() {
		parent::__construct(self::XS);
	}

	public function __output($type, $f) {
		if ($type == 'print')
			throw new Exception('XS attack!');
	}
};

/* ------------------------ output blocks --------------------------------- */
$echo_buf = '';
$echo_stack = array ();
function echo_begin () {
  global $echo_buf, $echo_stack;
  array_push ($echo_stack, strlen($echo_buf));
}

function echo_commit () {
  global $echo_buf, $echo_stack;
  array_pop ($echo_stack);
  echo '';
}

function echo_abort () {
  global $echo_buf, $echo_stack;
  $newlength = array_pop ($echo_stack);
  $echo_buf = substr($echo_buf, 0, $newlength);
}

function echo_filter ($type, $dst, $s) {
  /* 1) filter: $s -> $s2 
   * 2) call __output() on $s2's resulting policy
   * 3) return $s2;
   */
  global $echo_buf, $echo_stack;

  // XSS filtering code should go here.
  $s2 = $s;

  //perr ("ECHO\n");
  if ($mp = taint_get_policy ($s2)) {
    foreach ($mp as $ent) {
      $ent['policy']->__output ($type, $dst);
    }
  }

  if (sizeof($echo_stack) > 0) {
    //perr("BUFFERING\n");
    $echo_buf .= $s2;
    return '';
  } else {
    //perr("FLUSHING\n");
    $s2 = $echo_buf . $s2;
    $echo_buf = '';
    return $s2;
  }
}

/* ------------------------ directory integrity policy --------------------------------- */


?>
