--TEST--
TAINT merge the policies from two string zvals
--SKIPIF--
<?php if (!extension_loaded('taint')) die('skip no taint support'); ?>
--FILE--
<?php

class PolicyCollection {
  public $policies;

  function PolicyCollection ($policy=null) {
    if ($policy) {
      $this->policies = array ($policy);
    } else {
      $this->policies = array ();
    }
  }

  function addPolicy ($policy) {
    $this->policies[] = $policy;
  }

  function addPolicies ($list) {
    foreach ($list as $item) {
      $this->addPolicy ($item);
    }
  }

  function __toString () {
    return join (", ", $this->policies);
  }

  function __merge ($other_collection) {
    $merge_p = new PolicyCollection ();
    $merge_p->addPolicies ($this->policies);
    if ($other_collection) 
      $merge_p->addPolicies ($other_collection->policies);
    return $merge_p;
  }

  function __output($type, $obj) {
    foreach ($this->policies as $item) {
      $item->__output($type, $obj);
    }
  }
}

class Policy
{
  public $s;

  function Policy ($s) {
    $this->s = $s;
  }

  function __toString () {
    return $this->s;
  }

  function __output() {}
}

function print_policy ($obj) {
  echo "["; 
  echo taint_get_policy ($obj); 
  echo "]\n";
}

$pc_1 = new PolicyCollection (new Policy ("POLICY_1"));
$pc_2 = new PolicyCollection (new Policy ("POLICY_2"));
$pc_3 = new PolicyCollection (new Policy ("POLICY_3"));
$pc_4 = new PolicyCollection (new Policy ("POLICY_4"));
$pc_5 = new PolicyCollection (new Policy ("POLICY_5"));

$a = "aaa aaa aaa";
taint_set_policy ($a, $pc_1);
$b = "bbb bbb bbb";
taint_set_policy ($b, $pc_2);
$c = "ccc ccc ccc";
taint_set_policy ($c, $pc_3);
$d = ", ";
taint_set_policy ($d, $pc_4);
$e = "/usr/local/bin/foo";
taint_set_policy ($e, $pc_5);

echo "concat\n";
$x = $a . " " . $b;
echo $x; echo "\n";
print_policy ($a);
print_policy ($b);
print_policy ($x);

echo "concat2\n";
$x = "";
$x .= $a;
echo $x; echo "\n";
print_policy ($a);
print_policy ($x);

echo "wordwrap\n";
$x = wordwrap ($a, 3);
echo $x; echo "\n";
print_policy ($a);
print_policy ($x);

echo "strcspn\n";
$x = strcspn ($a, " ");
echo $x; echo "\n";
print_policy ($x);

echo "strspn\n";
$x = strspn ($a, "abc");
echo $x; echo "\n";
print_policy ($x);

echo "trim\n";
$a = "   aaa   ";
taint_set_policy ($a, $pc_1);
$x = trim ($a);
echo "[$x]\n";
print_policy ($x);
$x = ltrim ($a);
echo "[$x]\n";
print_policy ($x);
$x = rtrim ($a);
echo "[$x]\n";
print_policy ($x);

echo "implode\n";
$a = "aaa aaa aaa";
taint_set_policy ($a, $pc_1);
$x = implode (", ", array ($a, $b, $c));
echo "$x\n";
print_policy ($x);
$x = implode ($d, array ($a, $b, $c));
echo "$x\n";
print_policy ($x);

echo "explode\n";
$x = explode (" ", $a);
foreach ($x as $item) {
  echo "$item\n";
  print_policy ($item);
}

echo "strtok\n";
$item = strtok($a, " ");
while ($item !== false) {
  echo "$item\n";
  print_policy ($item);
  $item = strtok(" ");
}

echo "basename\n";
$x = basename ($e);
echo "$x\n";
print_policy ($x);

echo "dirname\n";
$x = dirname ($e);
echo "$x\n";
print_policy ($x);

echo "pathinfo\n";
$x = pathinfo ($e);
foreach ($x as $k => $item) {
  echo "$k => $item\n";
  print_policy ($k);
  print_policy ($item);
}

/*

echo "stristr\n";
$x = stristr ($a, "aaa");
echo "$x\n";
print_policy ($x);
*/

?>
--EXPECT--
concat
aaa aaa aaa bbb bbb bbb
[POLICY_1]
[POLICY_2]
[POLICY_1, POLICY_2]
concat2
aaa aaa aaa
[POLICY_1]
[POLICY_1]
wordwrap
aaa
aaa
aaa
[POLICY_1]
[POLICY_1]
strcspn
3
[]
strspn
3
[]
trim
[aaa]
[POLICY_1]
[aaa   ]
[POLICY_1]
[   aaa]
[POLICY_1]
implode
aaa aaa aaa, bbb bbb bbb, ccc ccc ccc
[POLICY_1, POLICY_2, POLICY_3]
aaa aaa aaa, bbb bbb bbb, ccc ccc ccc
[POLICY_1, POLICY_4, POLICY_2, POLICY_4, POLICY_3]
explode
aaa
[POLICY_1]
aaa
[POLICY_1]
aaa
[POLICY_1]
strtok
aaa
[POLICY_1]
aaa
[POLICY_1]
aaa
[POLICY_1]
basename
foo
[POLICY_5]
dirname
/usr/local/bin
[POLICY_5]
pathinfo
dirname => /usr/local/bin
[]
[POLICY_5]
basename => foo
[]
[POLICY_5]
filename => foo
[]
[POLICY_5]
