--TEST--
TAINT operations on doubles
--SKIPIF--
<?php if (!extension_loaded('taint')) die('skip no taint support'); ?>
--INI--
--COOKIE--
username=gotcha
--FILE--
<?php

function testop($str)
{
	$x55 = (double) 0x55;
	taint($x55, 0x55);
	$xaa = (double) 0xaa;
	taint($xaa, 0xaa);
	echo "$str -> ";
	eval("debug_zval_dump($str);");
}

echo <<< END
===== Notice =====
Some operators may convert one or more arguments and in the process
strip taint bits down to 0x1 (actually, whatever is configured with
taint_marks_non_str).  This is intentional, not a bug. In real life,
numbers will never have taint bits other than those configured with
taint_marks_non_str.

The && and || operators currently lose the taint bits of one operand
(they are implemented as ZEND_JMPZ_EX or ZEND_JMPNZ_EX which do
boolean evaluation of one operand, followed by ZEND_BOOL which does
boolean evaluation of the other operand).  Fixing this would require
changes to the code generator. Considering the low return for the
effort perhaps we should not try to maintain taint bits for booleans.

END;
echo "===== x55 =====\n";
testop('$x55');

echo "===== xaa =====\n";
testop('$xaa');

echo "===== arithmetic ops =====\n";
testop('$x55 + $xaa');
testop('$x55 - $xaa');
testop('$x55 * $xaa');
testop('$x55 / $xaa');
testop('$x55 % $xaa');
testop('-$x55');
testop('-$xaa');
testop('+$x55');
testop('+$xaa');
testop('++$x55');
testop('++$xaa');
testop('--$x55');
testop('--$xaa');
testop('$x55++');
testop('$xaa++');
testop('$x55--');
testop('$xaa--');

testop('$x55 += $xaa');
testop('$x55 -= $xaa');
testop('$x55 *= $xaa');
testop('$x55 /= $xaa');
testop('$x55 %= $xaa');

echo "===== bitwise ops =====\n";
testop('$x55 ^ $xaa');
testop('$x55 | $xaa');
testop('$x55 & $xaa');
testop('$x55 << $xaa');
testop('$x55 >> $xaa');
testop('~$x55');
testop('~$xaa');

testop('$x55 ^= $xaa');
testop('$x55 |= $xaa');
testop('$x55 &= $xaa');
testop('$x55 <<= $xaa');
testop('$x55 >>= $xaa');

echo "===== boolean ops =====\n";
testop('$x55 && $xaa');
testop('$xaa && $x55');
testop('$x55 || $xaa');
testop('$xaa || $x55');
testop('!$x55');
testop('!$xaa');

echo "===== relational ops =====\n";
testop('$x55 < $xaa');
testop('$x55 <= $xaa');
testop('$x55 == $xaa');
testop('$x55 != $xaa');
testop('$x55 === $xaa');
testop('$x55 !== $xaa');
testop('$x55 >= $xaa');
testop('$x55 > $xaa');

?>
--EXPECT--
===== Notice =====
Some operators may convert one or more arguments and in the process
strip taint bits down to 0x1 (actually, whatever is configured with
taint_marks_non_str).  This is intentional, not a bug. In real life,
numbers will never have taint bits other than those configured with
taint_marks_non_str.

The && and || operators currently lose the taint bits of one operand
(they are implemented as ZEND_JMPZ_EX or ZEND_JMPNZ_EX which do
boolean evaluation of one operand, followed by ZEND_BOOL which does
boolean evaluation of the other operand).  Fixing this would require
changes to the code generator. Considering the low return for the
effort perhaps we should not try to maintain taint bits for booleans.
===== x55 =====
$x55 -> double(85) refcount(2) taint(0x55)
===== xaa =====
$xaa -> double(170) refcount(2) taint(0xaa)
===== arithmetic ops =====
$x55 + $xaa -> double(255) refcount(1) taint(0xff)
$x55 - $xaa -> double(-85) refcount(1) taint(0xff)
$x55 * $xaa -> double(14450) refcount(1) taint(0xff)
$x55 / $xaa -> double(0.5) refcount(1) taint(0xff)
$x55 % $xaa -> long(85) refcount(1) taint(0x1)
-$x55 -> double(-85) refcount(1) taint(0x55)
-$xaa -> double(-170) refcount(1) taint(0xaa)
+$x55 -> double(85) refcount(1) taint(0x55)
+$xaa -> double(170) refcount(1) taint(0xaa)
++$x55 -> double(86) refcount(2) taint(0x55)
++$xaa -> double(171) refcount(2) taint(0xaa)
--$x55 -> double(84) refcount(2) taint(0x55)
--$xaa -> double(169) refcount(2) taint(0xaa)
$x55++ -> double(85) refcount(1) taint(0x55)
$xaa++ -> double(170) refcount(1) taint(0xaa)
$x55-- -> double(85) refcount(1) taint(0x55)
$xaa-- -> double(170) refcount(1) taint(0xaa)
$x55 += $xaa -> double(255) refcount(2) taint(0xab)
$x55 -= $xaa -> double(-85) refcount(2) taint(0xab)
$x55 *= $xaa -> double(14450) refcount(2) taint(0xab)
$x55 /= $xaa -> double(0.5) refcount(2) taint(0xab)
$x55 %= $xaa -> long(85) refcount(2) taint(0x1)
===== bitwise ops =====
$x55 ^ $xaa -> long(255) refcount(1) taint(0xff)
$x55 | $xaa -> long(255) refcount(1) taint(0xff)
$x55 & $xaa -> long(0) refcount(1) taint(0xff)
$x55 << $xaa -> long(87040) refcount(1) taint(0xff)
$x55 >> $xaa -> long(0) refcount(1) taint(0xff)
~$x55 -> long(-86) refcount(1) taint(0x55)
~$xaa -> long(-171) refcount(1) taint(0xaa)
$x55 ^= $xaa -> long(255) refcount(2) taint(0x1)
$x55 |= $xaa -> long(255) refcount(2) taint(0x1)
$x55 &= $xaa -> long(0) refcount(2) taint(0x1)
$x55 <<= $xaa -> long(87040) refcount(2) taint(0x1)
$x55 >>= $xaa -> long(0) refcount(2) taint(0x1)
===== boolean ops =====
$x55 && $xaa -> bool(true) refcount(1)
$xaa && $x55 -> bool(true) refcount(1) taint(0x1)
$x55 || $xaa -> bool(true) refcount(1) taint(0x1)
$xaa || $x55 -> bool(true) refcount(1)
!$x55 -> bool(false) refcount(1) taint(0x1)
!$xaa -> bool(false) refcount(1)
===== relational ops =====
$x55 < $xaa -> bool(true) refcount(1)
$x55 <= $xaa -> bool(true) refcount(1)
$x55 == $xaa -> bool(false) refcount(1)
$x55 != $xaa -> bool(true) refcount(1)
$x55 === $xaa -> bool(false) refcount(1)
$x55 !== $xaa -> bool(true) refcount(1)
$x55 >= $xaa -> bool(false) refcount(1)
$x55 > $xaa -> bool(false) refcount(1)
