--TEST--
TAINT basic SQL parser functions
--SKIPIF--
<?php if (!extension_loaded('taint')) die('skip no taint support'); ?>
--FILE--
<?php

require_once('taint/taint.php');

function test_query ($q, $print_tree=false) {
  echo $q . "\n";

  $tree = taint_parse_sql($q);
  if ($tree == NULL) { echo "Parse error?\n"; }

  if ($print_tree) {  print_r ($tree); }
  echo parse_tree_to_string ($tree) . "\n";
  return NULL;
}

test_query ("SET NAMES 'utf8'");

test_query ("INSERT INTO foo (v1, v2, v3) VALUES (111, 222, 333)");
test_query ("INSERT INTO foo (v1, v2, v3) VALUES ('x', 'y', 'z')");
test_query ("INSERT INTO foo (v1, v2, v3) VALUES (111, 222, 333), ('x', 'y', 'z')");
test_query ("INSERT INTO foo (v1, v2, v3) VALUES (111, '222', 'asdf')");
test_query ("INSERT INTO foo (v1, v2) VALUES (foo+1, bar)");
test_query ("INSERT INTO foo (v1, v2) SELECT foo, bar FROM table1");
test_query ("INSERT INTO foo (v1, v2) VALUES (foo+1, bar)");

test_query ("INSERT INTO foo (v1) SELECT foo FROM table1 ON DUPLICATE KEY UPDATE v1=v1");
test_query ("INSERT INTO foo (v1) VALUES (1) ON DUPLICATE KEY UPDATE v1=v1");

test_query ("SELECT foo, bar FROM table1 WHERE x=1");
test_query ("SELECT foo, bar FROM table1 WHERE x='y'");
test_query ("SELECT foo, bar FROM table1 WHERE x=y AND y=z");
test_query ("SELECT foo, 1, 'bar' FROM table1");
test_query ("SELECT foo, 1, 'bar' FROM table1, table2");
test_query ("SELECT foo, bar FROM table1 INNER JOIN table2 ON table1.x=table2.y");

test_query ("SELECT x FROM t1 WHERE x=1 AND y IN (SELECT z FROM t2)");
test_query ("SELECT max(foo), max(bar) FROM table1 WHERE x=1 GROUP BY table1.baz, table1.bazbaz");
test_query ("SELECT c1, c2 FROM t1 LEFT JOIN t2 ON (a=b) LEFT JOIN t3 ON (c=d)");

test_query ("SELECT c1 FROM t1 LEFT JOIN t2 AS nickname ON (a=b)");
test_query ("SELECT c1 FROM t1 LEFT JOIN (SELECT c2, c3 FROM t2) ON (x=y)");

test_query ("SELECT c1, c2 FROM t1 ORDER BY t1.c1, t1.c2 ASC, t1.c3 DESC");

test_query ("SELECT c1 FROM t1 JOIN t2 ON t1.x=t2.y");
test_query ("SELECT c1 FROM t1 JOIN t2 USING (x)");
test_query ("SELECT c1 FROM t1 JOIN t2 USING (x, y)");
test_query ("SELECT * FROM foo"); 
test_query ("SELECT -1 FROM t1");
test_query ("SELECT x, y, SUM( if (x=2, 2, y-1)) FROM t1");

test_query ("UPDATE t1 SET x=x+1, y=123 WHERE z=1");
test_query ("UPDATE t1 set x=x+1, y=123 WHERE z=1");

test_query ("DELETE FROM TopicInterest");
test_query ("DELETE FROM TopicInterest WHERE contactId=1");


//test_query ("select ContactInfo.contactId, firstName, lastName, PaperConflict.conflictType, PaperReview.reviewType, coalesce(preference, 0) as preference, group_concat(AllReviews.reviewType) as allReviews, coalesce(PaperTopics.topicInterestScore,0) as topicInterestScore from ContactInfo join PCMember using (contactId) left join PaperConflict on (PaperConflict.contactId=ContactInfo.contactId and PaperConflict.paperId=4) left join PaperReview on (PaperReview.contactId=ContactInfo.contactId and PaperReview.paperId=4) left join PaperReviewPreference on (PaperReviewPreference.contactId=ContactInfo.contactId and PaperReviewPreference.paperId=4) left join PaperReview as AllReviews on (AllReviews.contactId=ContactInfo.contactId) left join (select contactId, sum(if(interest=2,2,interest - 1)) as topicInterestScore from PaperTopic join TopicInterest using (topicId) where paperId=4 group by contactId) as PaperTopics on (PaperTopics.contactId=ContactInfo.contactId) group by email order by lastName, firstName, email");

//test_query ("insert into Paper SET title='foo', abstract='asdfasdfasdf', collaborators='', authorInformation='	foo		\n', blind=1, paperStorageId=1");
//test_query ("CREATE TEMPORARY TABLE Matches SELECT Paper.paperId paperId FROM Paper WHERE Paper.timeSubmitted>0 GROUP BY Paper.paperId");

//test_query ("SELECT PCMember.contactId, group_concat(reviewType separator '') FROM PCMember LEFT JOIN PaperReview ON (PCMember.contactId=PaperReview.contactId) GROUP BY PCMember.contactId");

// The following query should produce a parse error, but it does not
//test_query ("SELECT c1, c2 FROM t1 ORDER BY c1 GARBAGE;", true);

?>
--EXPECT--
SET NAMES 'utf8'
SET NAMES 'utf8'
INSERT INTO foo (v1, v2, v3) VALUES (111, 222, 333)
INSERT INTO foo (v1, v2, v3) VALUES (111, 222, 333)
INSERT INTO foo (v1, v2, v3) VALUES ('x', 'y', 'z')
INSERT INTO foo (v1, v2, v3) VALUES ('x', 'y', 'z')
INSERT INTO foo (v1, v2, v3) VALUES (111, 222, 333), ('x', 'y', 'z')
INSERT INTO foo (v1, v2, v3) VALUES (111, 222, 333), ('x', 'y', 'z')
INSERT INTO foo (v1, v2, v3) VALUES (111, '222', 'asdf')
INSERT INTO foo (v1, v2, v3) VALUES (111, '222', 'asdf')
INSERT INTO foo (v1, v2) VALUES (foo+1, bar)
INSERT INTO foo (v1, v2) VALUES ((foo + 1), bar)
INSERT INTO foo (v1, v2) SELECT foo, bar FROM table1
INSERT INTO foo (v1, v2) SELECT foo, bar FROM table1
INSERT INTO foo (v1, v2) VALUES (foo+1, bar)
INSERT INTO foo (v1, v2) VALUES ((foo + 1), bar)
INSERT INTO foo (v1) SELECT foo FROM table1 ON DUPLICATE KEY UPDATE v1=v1
INSERT INTO foo (v1) SELECT foo FROM table1 ON DUPLICATE KEY UPDATE v1=v1
INSERT INTO foo (v1) VALUES (1) ON DUPLICATE KEY UPDATE v1=v1
INSERT INTO foo (v1) VALUES (1) ON DUPLICATE KEY UPDATE v1=v1
SELECT foo, bar FROM table1 WHERE x=1
SELECT foo, bar FROM table1 WHERE (x = 1)
SELECT foo, bar FROM table1 WHERE x='y'
SELECT foo, bar FROM table1 WHERE (x = 'y')
SELECT foo, bar FROM table1 WHERE x=y AND y=z
SELECT foo, bar FROM table1 WHERE ((x = y) AND (y = z))
SELECT foo, 1, 'bar' FROM table1
SELECT foo, 1, 'bar' FROM table1
SELECT foo, 1, 'bar' FROM table1, table2
SELECT foo, 1, 'bar' FROM table1, table2
SELECT foo, bar FROM table1 INNER JOIN table2 ON table1.x=table2.y
SELECT foo, bar FROM table1 INNER JOIN table2 ON (table1.x = table2.y)
SELECT x FROM t1 WHERE x=1 AND y IN (SELECT z FROM t2)
SELECT x FROM t1 WHERE ((x = 1) AND (y IN (SELECT z FROM t2)))
SELECT max(foo), max(bar) FROM table1 WHERE x=1 GROUP BY table1.baz, table1.bazbaz
SELECT max(foo), max(bar) FROM table1 WHERE (x = 1) GROUP BY table1.baz, table1.bazbaz
SELECT c1, c2 FROM t1 LEFT JOIN t2 ON (a=b) LEFT JOIN t3 ON (c=d)
SELECT c1, c2 FROM t1 LEFT JOIN t2 ON (a = b) LEFT JOIN t3 ON (c = d)
SELECT c1 FROM t1 LEFT JOIN t2 AS nickname ON (a=b)
SELECT c1 FROM t1 LEFT JOIN t2 AS nickname ON (a = b)
SELECT c1 FROM t1 LEFT JOIN (SELECT c2, c3 FROM t2) ON (x=y)
SELECT c1 FROM t1 LEFT JOIN (SELECT c2, c3 FROM t2) ON (x = y)
SELECT c1, c2 FROM t1 ORDER BY t1.c1, t1.c2 ASC, t1.c3 DESC
SELECT c1, c2 FROM t1 ORDER BY t1.c1, t1.c2 ASC, t1.c3 DESC
SELECT c1 FROM t1 JOIN t2 ON t1.x=t2.y
SELECT c1 FROM t1 JOIN t2 ON (t1.x = t2.y)
SELECT c1 FROM t1 JOIN t2 USING (x)
SELECT c1 FROM t1 JOIN t2 USING (x)
SELECT c1 FROM t1 JOIN t2 USING (x, y)
SELECT c1 FROM t1 JOIN t2 USING (x, y)
SELECT * FROM foo
SELECT * FROM foo
SELECT -1 FROM t1
SELECT (- 1) FROM t1
SELECT x, y, SUM( if (x=2, 2, y-1)) FROM t1
SELECT x, y, SUM(if((x = 2), 2, (y - 1))) FROM t1
UPDATE t1 SET x=x+1, y=123 WHERE z=1
UPDATE t1 SET x=(x + 1), y=123 WHERE (z = 1)
UPDATE t1 set x=x+1, y=123 WHERE z=1
UPDATE t1 SET x=(x + 1), y=123 WHERE (z = 1)
DELETE FROM TopicInterest
DELETE FROM TopicInterest
DELETE FROM TopicInterest WHERE contactId=1
DELETE FROM TopicInterest WHERE (contactId = 1)
