<?php

/* Microbenchmarks
 *  1) concat
 *  2) substr ?
 *  3) echo through filter?
 * Memory footprint
 * Page generation time. CPU bound 
 */

$stderr = fopen ("php://stderr", 'w');
function perr2 ($s) {
  global $stderr;
  fwrite ($stderr, $s);
  fflush($stderr);
}

function usage () {
  perr2("main.php reps {-u | -n | -p}\n");
  perr2("  reps is the number of iterations\n");
  perr2("  -u use with unmodified php\n");
  perr2("  -n no policies (use with unmodified php)\n");
  perr2("  -p dummy policies (use with modified php)\n");
  exit ();
}

if ($argc < 3) {
  perr2("not enough args\n");
  usage ();
}
$specific_tests = NULL;
if ($argc > 3) {
  $specific_tests = array_slice ($argv, 3);
}

$reps = $argv[1];

$a = "A";
$b = "B";
$x = 1;
$y = 2;

if ($argv[2] == '-u') {
  $mode = "unmodified";
} else if ($argv[2] == '-n') {
  $mode = "nopolicy";
} else if ($argv[2] == '-p') {
  $mode = "policy";
  taint_set_policy ($a, new DummyPolicy ("pa"));
  taint_set_policy ($b, new DummyPolicy ("pb"));
  taint_set_policy ($x, new DummyPolicy ("px"));
  taint_set_policy ($y, new DummyPolicy ("py"));
  
} else {
  echo "must specify -u, -n or -p\n";
  usage ();
}

function null_filter() {
}

if (!$specific_tests || in_array("filter", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    test_null_filter($a);
  }
  $end = gettimeofday (true);
  print_it ("filter", $reps, $start, $end);
}

function print_it ($s, $reps, $start, $end) {
  $usec = (($end - $start) / $reps) * 1000000;
  perr2 (sprintf ("%-20s %8d %10.3f\n", $s, $reps, $usec));
}

if (!$specific_tests || in_array("concat", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    $c = $a . $b;
  }
  $end = gettimeofday (true);
  print_it ("concat", $reps, $start, $end);
}

if (!$specific_tests || in_array("substr", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    $c = substr($a, 0, 2);
  }
  $end = gettimeofday (true);
  print_it ("substr", $reps, $start, $end);
}

if (!$specific_tests || in_array("htmlentities", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    $c = htmlentities($a);
  }
  $end = gettimeofday (true);
  print_it ("htmlentities", $reps, $start, $end);
}

if (!$specific_tests || in_array("add", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    $z = $x + $y;
  }
  $end = gettimeofday (true);
  print_it ("add", $reps, $start, $end);
}

if (!$specific_tests || in_array("subtract", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    $c = $x - $y;
  }
  $end = gettimeofday (true);
  print_it ("subtract", $reps, $start, $end);
}

if (!$specific_tests || in_array("assign", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    $tmp = $a;
  }
  $end = gettimeofday (true);
  print_it ("assign", $reps, $start, $end);
}

function f0 () {
  return;
}

if (!$specific_tests || in_array("function_noargs", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    f0();
  }
  $end = gettimeofday (true);
  print_it ("function_noargs", $reps, $start, $end);
}

function f1 ($arg) {
  return;
}

if (!$specific_tests || in_array("function_1arg", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    f1($a);
  }
  $end = gettimeofday (true);
  print_it ("function_1arg", $reps, $start, $end);
}

if (!$specific_tests || in_array("echo1", $specific_tests)) {
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    echo $a;
  }
  $end = gettimeofday (true);
  print_it ("echo1", $reps, $start, $end);
}

if (!$specific_tests || in_array("echo1000", $specific_tests)) {
  $s = '';
  for ($i=0; $i<1000; $i++)
    $s .= $a;
  
  $start = gettimeofday (true);
  for ($i=0; $i<$reps; $i++) {
    echo $s;
  }
  $end = gettimeofday (true);
  print_it ("echo1000", $reps, $start, $end);
}

if (!$specific_tests || in_array("fileopen", $specific_tests)) {
  $batch = 10;
  $s = str_repeat('X', 1000); // no policy on data, just container

  for ($i=0; $i<$batch; $i++) {
    file_put_contents("/tmp/test-file$i.txt", $s);
    if ($mode == 'policy') {
      $f = fopen ("/tmp/test-file$i.txt", 'r');
      taint_set_file_policy ($f, new DummyPolicy ("fp"));
      fclose ($f);
    }
  }

  $cum_time = 0;
  $i=0;
  while ($i < $reps) {
    $fh = array ();
    $start = gettimeofday (true);
    for ($j=0; $j<$batch && $i < $reps; $j++) {
      $fh[] = fopen ("/tmp/test-file$j.txt", 'r');
      $i++;
    }
    $end = gettimeofday (true);
    $cum_time += ($end - $start);

    foreach ($fh as $h)
      fclose ($h);
  }

  for ($i=0; $i<$batch; $i++)
    unlink("/tmp/test-file$i.txt");

  print_it ("fileopen", $reps, 0, $cum_time);
}

if (!$specific_tests || in_array("fileread", $specific_tests)) {
  $batch = 10;
  $s = '';
  for ($i=0; $i<1000; $i++)
    $s .= $a;

  for ($i=0; $i<$batch; $i++)
    file_put_contents("/tmp/test-file$i.txt", $s);

  $start = gettimeofday (true);
  for ($i=0; $i < $reps; $i++) {
    $x = file_get_contents("/tmp/test-file" . $i % $batch. ".txt", $s);
    //assert (strlen($x) == 1000);
  }
  $end = gettimeofday (true);

  for ($i=0; $i<$batch; $i++)
    unlink("/tmp/test-file$i.txt");

  print_it ("fileread", $reps, $start, $end);
}

if (!$specific_tests || in_array("filewrite", $specific_tests)) {
  $batch = 10;
  $s = '';
  for ($i=0; $i<1000; $i++)
    $s .= $a;
  $cum_time = 0;
  $i=0;
  while ($i < $reps) {
    $start = gettimeofday (true);
    $max = 0;
    for ($j=0; $j<$batch && $i < $reps; $j++) {
      file_put_contents("/tmp/test-file$j.txt", $s);
      $i++;
      $max++;
    }
    $end = gettimeofday (true);
    $cum_time += ($end - $start);

    for ($j=0; $j<$max; $j++) {
      unlink("/tmp/test-file$j.txt");
    }
  }
  print_it ("filewrite", $reps, 0, $cum_time);
}

/* DB stuff ------------------ */
function query ($q, $link) {
  $r = mysql_query ($q, $link);
  if (!$r) {
    perr2("mysql error: " . mysql_error () . "\n");
    perr2(" query: $q\n");
  }
  return $r;
}

function db_loop($name, $reps, $q, $where, $doquery, $link) {
  global $mode;
  $start = gettimeofday (true);
  
  for ($i=0; $i<$reps; $i++) {
    $_q = $where ? $q . " WHERE id=" . ($i+1) : $q;
    //perr2("query: $_q\n");
    //perr2("  policy: " . mp_to_str(taint_get_policy($q)) . "\n");
    if ($mode == 'nopolicy' || $mode == 'policy') {
      //stats_begin("translate");
      $_q = sql_translate_express($_q);
      //stats_end("translate");
    }
    //perr2("query: $_q\n");
    //perr2(mp_to_str(taint_get_policy ($_q)) . "\n");

    if ($doquery) {
      //stats_begin("query");
      $r = mysql_query ($_q, $link);
      //stats_end("query");
      if (!$r) {
        perr2("mysql error: " . mysql_error () . "\n");
        perr2(" query: $q\n");
      }
      if (is_resource($r)) {
        assert (mysql_num_rows($r) == 1);
        //stats_begin("fetch");
        $row = ($mode == 'unmodified') ? mysql_fetch_row($r) : resin_mysql_fetch_row($r);
        //stats_end("fetch");
      }
    }
  }
  $end = gettimeofday (true);
  print_it ($name, $reps, $start, $end);
}

function db_test ($name, $reps, $cols, $val, $doquery) {
  global $mode;
  if ($doquery) {
    $DB = "testdb";
    $link = mysql_connect ();
    assert ($link);

    mysql_query ("DROP DATABASE $DB", $link);

    $r = query ("CREATE DATABASE $DB", $link);
    assert ($r);
    mysql_select_db ($DB, $link);
  }
    
  $coldef = array ();
  for ($i=0; $i<$cols; $i++)
    $coldef[] = "c$i INT";
  if ($mode == "nopolicy" || $mode == "policy")
    for ($i=0; $i<$cols; $i++)
      $coldef[] = "c$i" . "_policy BLOB";
      //$coldef[] = "c$i" . "_policy VARCHAR(512)";

  $insertcols = array ();
  $insertvals = array ();
  for ($i=0; $i<$cols; $i++) {
    $insertcols[] = "c$i";
    $insertvals[] = "$val";
  }

  if ($doquery) {
    $q = "CREATE TABLE foo (id INT NOT NULL UNIQUE KEY AUTO_INCREMENT, " . implode(',', $coldef) . ")";
    $r = query ($q, $link);
    echo mysql_error();
    assert ($r);
  }

  //stats_reset ();
  $q = "INSERT INTO foo (" . implode(',', $insertcols) . ") VALUES (" . implode(',', $insertvals) . ")";
  db_loop($name . "_insert", $reps, $q, false, $doquery, $link);
  //stats_report ();

  //stats_reset ();
  $q = "SELECT " . implode(',', $insertcols) . " FROM foo";
  db_loop($name . "_select", $reps, $q, true, $doquery, $link);
  //stats_report ();

  //stats_reset ();
  $q = "DELETE FROM foo";
  db_loop($name . "_delete", $reps, $q, true, $doquery, $link);
  //stats_report ();

  if ($doquery) {
    $r = query ("DROP DATABASE $DB", $link);
    assert ($r);
    mysql_close($link);
  }
}

if (!$specific_tests || in_array("sqltrans1", $specific_tests)) {
  db_test ("sqltrans1", $reps, 1, $x, 0);
}

if (!$specific_tests || in_array("sqltrans10", $specific_tests)) {
  db_test ("sqltrans10", $reps, 10, $x, 0);
}

if (!$specific_tests || in_array("sqlquery1", $specific_tests)) {
  db_test ("sqlquery1", $reps, 1, $x, 1);
}

if (!$specific_tests || in_array("sqlquery6", $specific_tests)) {
  db_test ("sqlquery6", $reps, 6, $x, 1);
}

if (!$specific_tests || in_array("sqlquery10", $specific_tests)) {
  db_test ("sqlquery10", $reps, 10, $x, 1);
}


?>
